// Author: Marek Handl
// Date: February 7, 2008
// CTU, FEE, X36PAA
// SAT (Boolean satisfiability problem) problem - solution using simulated annealing

// see class Instance for the simulated annealing algorithm
// you can find the set up of initialTemperature and other parameters as well as the logic there

// now the computation on the same data is repeated INSTANCESCOUNT times to get more represantive statistics
// about the program, but this repetitive performance has no effect on the solution returned

// normally you should use INSTANCESCOUNT = 1. The iterations are not interconnected, you will not
// get the best solution from all the iterations, but just from one.

#include <vector>
#include <map>
#include <sstream>
#include <iostream>
#include <string>
#include <fstream>
#include "assert.h"
#include "process.h"
#include "windows.h"
#include <algorithm>
#include <math.h>
#include <time.h>

using namespace std;

#define USE_TESTBEST true	// enable/disable algorithm improvement, that always returns the best solution found
#define ITERATIONSCOUNT 5	// how many times to repeat the computation (to mitigate random effect), default is 5
#define SETALL false  		// defines starting state (true means all variables will be set to 1 at the beginning)

// class declarations
class Instance;
class Formula;
class Term;
class Variable;
class Evaluation;

// *********************************************************************************
// *********************************************************************************
class Evaluation {
	map<int, bool> values;
	float cost;
public:
	Evaluation() {
		cost = -1.0;	// sentinel
	}

	bool get(const int key) const {
		assert(values.find(key) != values.end());
		return values.find(key)->second;
	}
	void set(const int key) {
		values[key] = true;
		cost = -1.0;		// make cost value invalid
	}
	void unset(const int key) {
		values[key] = false;
		cost = -1.0;		// make cost value invalid
	}
	// change value of the variable to the opposite value
	void flip(const int key) {
		assert(values.find(key) != values.end());
		values[key] = (values.find(key)->second ? false : true);
		cost = -1.0;		// make cost value invalid
	}

	void setAllTrue() {
		map<int, bool>::iterator iter;
		for (iter = values.begin(); iter != values.end(); ++iter) {
			iter->second = true;
		}
		cost = -1.0;		// make cost value invalid
	}
	void setAllFalse() {
		map<int, bool>::iterator iter;
		for (iter = values.begin(); iter != values.end(); ++iter) {
			iter->second = false;
		}
		cost = -1.0;		// make cost value invalid
	}
	// asignment operator
	Evaluation& operator=(const Evaluation &eval) {
		map<int, bool>::const_iterator iter;
		for (iter = eval.values.begin(); iter != eval.values.end(); ++iter) {
			values[iter->first] = iter->second;
		}
		cost = eval.getCost();
		return *this;
	}
	// equals operator
	bool operator==(const Evaluation &eval) const {
		map<int, bool>::const_iterator iter;
		for (iter = values.begin(); iter != values.end(); ++iter) {
			if (iter->second != eval.get(iter->first)) {
				return false;
			}
		}
		return true;
	}
	
	float getCost() const {
		return cost;
	}
	void setCost(const float c) {
		cost = c;
	}

	// print evaluation to a stream
	void printResult(ostream &out, const int variablesCount) const {
		for (int i=1; i <= variablesCount; ++i) {
			out << static_cast<int>(values.find(i)->second) << " ";
		}
	}
};

// *********************************************************************************
// *********************************************************************************
class Variable {
	int id;
	bool negative;

public:
	Variable(const int i, const bool n) : id(i), negative(n) {}

	bool isSatisfied(const Evaluation &evaluation) const {
		return evaluation.get(id) ^ negative;	// xor
	}
	void outputToFile(ofstream &fout, const Formula &formula) const;
};

// *********************************************************************************
// *********************************************************************************
class MaxTerm {
	vector<Variable> variables;
public:
	void addVariable(const Variable &var) {
		variables.push_back(var);
	}
	void erase() {
		variables.clear();
	}
	void outputToFile(ofstream &fout, const Formula& formula) const {
		vector<Variable>::const_iterator iter;
		for (iter = variables.begin(); iter != variables.end(); ++iter) {
			iter->outputToFile(fout, formula);
		}
	}
	bool isSatisfied(const Evaluation &evaluation) const {
		vector<Variable>::const_iterator iter;
		for (iter = variables.begin(); iter != variables.end(); ++iter) {
			if (iter->isSatisfied(evaluation)) {
				return true;	// first satisfied variable makes it the whole maxterm satisfied
			}
		}
		return false;
	}
};

// *********************************************************************************
// *********************************************************************************
class Formula {
	vector<MaxTerm> terms;
	map<int, int> weights;
	int variablesCount, termsCount, variablesInTerm, maxAchievableWeight;

public:
	Formula() {
		maxAchievableWeight = 0;
	}
	void reset() {
		terms.clear();
		weights.clear();
		maxAchievableWeight = 0;
	}
	void addTerm(const MaxTerm &term) {
		terms.push_back(term);
	}
	void setCounts(const int vc, const int tc, const int vit) {
		variablesCount = vc;
		termsCount = tc;
		variablesInTerm = vit;
	}
	int getVariablesCount() const {
		return variablesCount;
	}
	int getTermsCount() const {
		return termsCount;
	}
	int getVariablesInTerm() const {
		return variablesInTerm;
	}
	// generate random weights for variables
	void generateWeights(const int maxWeight) {
		map<int, int>::iterator iter;
		for (iter = weights.begin(); iter != weights.end(); ++iter) {
			iter->second = static_cast<int>(1 + static_cast<float>(rand())/RAND_MAX*(maxWeight-1));	// 1 .. maxWeight
			assert(iter->second > 0 && iter->second <= maxWeight);
		}
	}
	// used to create new input file containing generated weights as well
	void saveInputFile(ofstream &fout) const {
		vector<MaxTerm>::const_iterator iter;
		// print out basic parameters
		fout << variablesCount << " " << termsCount << " " << variablesInTerm << endl;
		// print out terms
		for (iter = terms.begin(); iter != terms.end(); ++iter) {
			iter->outputToFile(fout, *this);
			fout << endl;
		}
		// print out weights
		for (int i=1; i<variablesCount; ++i) {
			fout << getWeight(i) << " ";
		}
		fout << endl;
	}

	// get weight of a variable
	int getWeight(const int key) const {
		assert(weights.find(key) != weights.end());
		return weights.find(key)->second;
	}

	// set weight for a variable
	void setWeight(const int key, const int weight) {
		if (weights.find(key) != weights.end()) {	// if there already is a value with the key
			maxAchievableWeight -= weights.find(key)->second;
		}
		weights[key] = weight;
		maxAchievableWeight += weights.find(key)->second;
	}

	// get sum of weights of all variables
	int getMaxAchievableWeight() const {
		return maxAchievableWeight;
	}

	// sum weights of all variables in the formula that are true in the evaluation
	int weightOfTrueVariables(const Evaluation &evaluation) const {
		map<int, int>::const_iterator iter;
		int sumWeight = 0;
		for (iter = weights.begin(); iter != weights.end(); ++iter) {
			if (evaluation.get(iter->first)) {
				sumWeight += iter->second;
			}
		}
		return sumWeight;
	}

	// how many terms are satisfied with the evaluation
	int countTrueTerms(const Evaluation &evaluation) const {
		vector<MaxTerm>::const_iterator iter;
		int count = 0;
		for (iter = terms.begin(); iter != terms.end(); ++iter) {
			if (iter->isSatisfied(evaluation)) {
				++count;
			}
		}
		return count;
	}
};

// *********************************************************************************
// *********************************************************************************
class Instance {
	Formula *formula;
	Evaluation *evaluation, best, bestSolution;
	float temperature, minTemperature, cooldownFactor, costFactor, improvementOfTestBest;
	int equilibrium, solutionsCount, worstWeight, statesTested;
	bool wasNotBest;

public:
	Instance(Formula &f, Evaluation &e, const float initT, const float minT, const float cdf, const int eqf, const float cf) {
		wasNotBest = false;
		improvementOfTestBest = 0.0;
		worstWeight = f.getMaxAchievableWeight()+1;	// every solution found will be worse
		solutionsCount = 0;
		statesTested = 0;
		formula = &f;
		evaluation = &e;
		minTemperature = minT;
		cooldownFactor = cdf;
		equilibrium = eqf * formula->getVariablesCount();
		costFactor = cf;
		// INITIAL TEMPERATURE FOR SIMULATED ANNEALING
		// the idea is to have the probability that a worse state is chosen equal to 1/2
		// 1/2 == exp(-delta / T)
		// T = delta / ln(2)
		// delta is maximally (costFactor + 1)
		temperature = initT * (costFactor+1) / static_cast<float>(log(2.0));
	}

	// SIMULATED ANNEALING MAIN LOOP
	int solve() {
		int steps;
		best = *evaluation;		// initialize best
		if (isCorrectSolution()) {	// initialize number of solutions and worst solution weight
			++solutionsCount;
			worstWeight = formula->weightOfTrueVariables(*evaluation);
			bestSolution = *evaluation;
		}

		while (temperature > minTemperature) {		// i.e. while not frozen
			steps = 0;
			while (steps < equilibrium) {		// i.e. while not equilibrium
				++steps;
				if (changeState() && getCost(*evaluation) > getCost(best)) {	// find a new state or keep the same
					best = *evaluation;
				}
			}
			temperature *= cooldownFactor;		// i.e. cool down
			statesTested += steps;
		}
		*evaluation = best;

		// if "testbest" improvement is enabled
		// it will compare the current solution with the best found so far (based on weight sum)
		// if the current is different, discard it and keep the best one
		// note that the current state might not be a solution, in that case whatever is stored in bestSolution is better, even though it might have a worse weight sum
		// that is the improvementOfTestBest will be negative in those cases
		if (USE_TESTBEST) {
			if (!isCurrentBest()) {
				assert(!wasNotBest);
				wasNotBest = true;
				improvementOfTestBest = static_cast<float>(formula->weightOfTrueVariables(bestSolution) - formula->weightOfTrueVariables(*evaluation))*100 / formula->weightOfTrueVariables(*evaluation);
				*evaluation = bestSolution;
				assert(isCorrectSolution());
			}
		}

		// if none solution is found, the worstWeight still contains implicit value (which is more than the theoretical max)
		// change it to the current weight, so that the result is not confusing
		if (worstWeight > formula->getMaxAchievableWeight()) {
			worstWeight = formula->weightOfTrueVariables(*evaluation);
		}
		return statesTested;
	}

	float getCost(Evaluation &eval) const {
		if (eval.getCost() < 0) {
			// COST FUNCTION FOR SIMULATED ANNEALING
			// costFactor * countOfTrueTerms / countOfAllTerms   +   weightOfTrueVariables / weightOfAllVariables
			// value is from 0 to (costFactor+1)
			float cost = costFactor * static_cast<float>(formula->countTrueTerms(eval)) / formula->getTermsCount() + static_cast<float>(formula->weightOfTrueVariables(eval)) / formula->getMaxAchievableWeight();
			eval.setCost(cost);
		}
		return eval.getCost();
	}

	// find a neighbor and take it or keep the current state (based on probability)
	bool changeState() {
		bool change = false;
		int newWorstWeight;
		float oldCost = getCost(*evaluation);	// remember cost of the current state

		// get random variable
		int index = 1 + rand() * (formula->getVariablesCount()-1) / RAND_MAX ;
		assert(index >= 1 && index <= formula->getVariablesCount());
		evaluation->flip(index);	// move to a new state

		// compare new state with the previous
		// if the new state is better
		if (getCost(*evaluation) > oldCost) {	
			// keep the new state
			change = true;
		}
		// if the new state is worse, but random value is less than threshold
		else if (static_cast<float>(rand())/RAND_MAX < exp(-oldCost-getCost(*evaluation)/temperature)) {
			// keep the new state
			change = true;
		}
		// if the state is worse and probability is against us
		else {
			evaluation->flip(index);	// take the change back
			evaluation->setCost(oldCost);
		}
		
		
		// keep track of the number of real solutions (formula satisfied) found so far
		if (change && isCorrectSolution()) {
			++solutionsCount;
			// keep track of the best solution found (based on weight sum)
			if ((bestSolution.getCost() < 0) || formula->weightOfTrueVariables(*evaluation) > formula->weightOfTrueVariables(bestSolution)) {
				bestSolution = *evaluation;
			}
			// keep track of the worst solution found (based on weight sum)
			if (worstWeight > (newWorstWeight = formula->weightOfTrueVariables(*evaluation))) {
				worstWeight = newWorstWeight;
			}
		}

		return change;
	}

	// is the current solution the best solution found?
	bool isCurrentBest() const {
		return (bestSolution.getCost() < 0 || bestSolution == *evaluation);		// uninitialized bestSolution will return true
	}

	// is the solution (defined by evaluation) a correct solution (i.e. is the formula satisfied?)
	bool isCorrectSolution() const {
		return (formula->countTrueTerms(*evaluation) == formula->getTermsCount());
	}

	// print info about the computation and result
	void printResult(ostream &out) const {
		out << endl;
		out << "Number of solutions found: " << solutionsCount << endl;
		out << "States tested: " << statesTested << endl;
		out << "Result weight: " << formula->weightOfTrueVariables(*evaluation) << endl;
		out << "Worst weight: " << worstWeight << endl;
		out << "Evaluation of variables: " << endl;
		printSolution(out);
	}

	// print result
	void printSolution(ostream &out) const {
		if (!isCorrectSolution()) {
			out << "# NOT A SOLUTION - formula not fully satisfied!!! (" << formula->countTrueTerms(*evaluation) << " out of " << formula->getTermsCount() << " terms satisfied)" << endl;
		}
		out << "# weight = " << formula->weightOfTrueVariables(*evaluation) << endl;
		evaluation->printResult(out, formula->getVariablesCount());
		out << endl;
	}

	int getSolutionsCount() const {
		return solutionsCount;
	}
	int getWeight() const {
		return formula->weightOfTrueVariables(*evaluation);
	}
	int getWorstWeight() const {
		return worstWeight;
	}
	int getSatisfiedTerms() const {
		return formula->countTrueTerms(*evaluation);
	}
	float getImprovementRatio() const {
		assert(worstWeight <= getWeight());
		if (worstWeight == 0) {
			return 0;
		}
		return static_cast<float>(getWeight() - worstWeight) / worstWeight;
	}
	bool foundBest() const {
		return !wasNotBest;
	}
	float getImprovementOfTestBest() const {
		return improvementOfTestBest;
	}
};

// *********************************************************************************
void Variable::outputToFile(ofstream &fout, const Formula &formula) const {
	fout << (this->negative ? "-" : "") << this->id << " ";
}

// *********************************************************************************
void readFromFile(ifstream &fin, Formula &formula, Evaluation &evaluation, bool withWeights = false) {
	string line;
	istringstream is;
	int variablesCount, termsCount, variablesInTerm, variable;
	int lineCount=0, inLineCount=0, weight = 0;
	MaxTerm term;

	try {		
		if (fin.rdstate() & fstream::failbit) {		// if file not found or unable to read from it
			throw "Input file not found";
		}
		while (getline(fin, line) && (line.length() == 0 || line.at(0) == '#'));	// skip comments and empty lines
		
		// read basic parameters
		is.str(line);
		is >> variablesCount >> termsCount >> variablesInTerm;
		formula.setCounts(variablesCount, termsCount, variablesInTerm);
		
		// read terms
		for (lineCount = 0; lineCount < termsCount; ++lineCount) {
			getline(fin, line);
			while (line == "" || line.at(0) == '#') {
				getline(fin, line);	// skip comments and empty lines
			}			
			term.erase();
			is.clear();
			is.str(line);
			is >> variable;
			inLineCount = 0;
			while (variable != 0) {		// either read all variablesInTerm values or stop once you read a zero
				if (++inLineCount > variablesInTerm) {
					break;
				}
				term.addVariable(Variable(abs(variable), variable < 0));
				evaluation.set(abs(variable));
				formula.setWeight(abs(variable), 0);	// implicit weight is 0
				is >> variable;				
			}
			formula.addTerm(term);
		}

		// read weights
		if (withWeights) {
			getline(fin, line);
			is.clear();
			is.str(line);
			for (int i=1; i<=variablesCount; ++i) {
				is >> weight;
				formula.setWeight(i, weight);
				evaluation.get(i);	// just a check that all variables are used in terms (if not, an assertion will be thrown)
			}
		}
	}
	catch (exception &e) {
		fin.close();
		cout << "Input file error: " << e.what() << endl;
		exit(1);
	}
}

// dump formula with generated weights into a file
void saveInputFile(ofstream &fout, Formula &formula) {
	fout << "# first line syntax: variablesCount termsCount maxVariablesInTerm" << endl;
	fout << "# last line syntax: weight weight weight ... (weight of each variable)" << endl;
	fout << "# syntax of the other lines: variable variable variable ... (each line represents a maxterm)" << endl;
	fout << "# you can use 0 as a variable if there are less variables than maxVariablesInTerm" << endl;
	fout << "# variables and weights are positive integers" << endl;
	formula.saveInputFile(fout);
}


// *********************************************************************************
// *********************************************************************************
int main(int argc, char* argv[]) {
	ifstream fin, fileDataIn;
	ofstream fout;
	string fileNameIn = "input.txt", fileNameOutNewInput = "input_ww.txt", fileNameOut = "output.txt";
	bool generateInputFile = false;
	int maxWeight, equilibriumFactor;
	float initialTemperatureFactor, minTemperature, cooldownFactor, costFactor;
	Formula formula;
	Evaluation evaluation;

	// usage instructions
	if (argc < 2 || (string(argv[1]) == "-gw" && argc < 3)) {
		cout << "Usage instructions" << endl;
		cout << "To generate a new input file with weights from an existing file without weights use:" << endl;
		cout << "   sat.exe -gw inputFile maxWeight" << endl;
		cout << "The new file will be named ww_inputFile." << endl << endl;
		cout << "To find solution of a SAT problem use:" << endl;
		
		cout << "   sat.exe inputFile" << endl;
		cout << "   where inputFile looks like this:" << endl;
		cout << "     initialTemperatureFactor minimalTemperature cooldownFactor equilibriumFactor costFactor" << endl;
		cout << "     inputFile1" << endl;
		cout << "     inputFile2" << endl;
		cout << "     ..." << endl;
		exit(1);
	}

	srand((unsigned int)time((time_t *)NULL));	// initialiaze random number generator

	if (string(argv[1]) == "-gw") {		// just generating an input file with weights
		fileNameIn = string(argv[2]);
		maxWeight = atoi(argv[3]);
		fileNameOutNewInput = "ww_"+fileNameIn;

		fin.open(fileNameIn.c_str(), fstream::in);
		readFromFile(fin, formula, evaluation);
		formula.generateWeights(maxWeight);
		fout.open(fileNameOutNewInput.c_str(), fstream::out);
		saveInputFile(fout, formula);	
		fin.close();
		fout.close();
		return 0;	// work finished
	}

	else {	// do a computation
		string line;
		vector<string> inputFiles;
		istringstream is;
		fileNameIn = string(argv[1]);
		fin.open(fileNameIn.c_str(), fstream::in);
		
		// read basic parameters
		try {
			getline(fin, line);
			is.str(line);
			is >> initialTemperatureFactor >> minTemperature >> cooldownFactor >> equilibriumFactor >> costFactor;
			// read filenames
			while (getline(fin, line)) {
				inputFiles.push_back(line);
			}
		}
		catch (exception e) {
			fin.close();
			cout << "Input file (" << fileNameIn << ") error: " << e.what();
			exit(1);
		}
		fin.close();

		if (inputFiles.size() < 1) {
			cout << "No data files specified" << endl;
			exit(1);
		}

		// do a computation for each input file
		int statesTested = 0, solutionsCount = 0, iterations = 0, solutionNotFound = 0;
		float satisfiedTerms = 0.0, worstWeight = 0.0, improvementTestBest = 0.0;
		vector<string>::const_iterator iter;
		for (iter = inputFiles.begin(); iter != inputFiles.end(); ++iter) {
			fileDataIn.open(iter->c_str(), fstream::in);
			fout.open(((*iter)+".res").c_str(), fstream::out);
			formula.reset();
			readFromFile(fileDataIn, formula, evaluation, true);
			fileDataIn.close();

			for (int i=0; i < ITERATIONSCOUNT; ++i) {	// repeat the same instance a couple of times to mitigate random effect
				// get some evaluation - starting state
				if (SETALL) {
					evaluation.setAllTrue();
				}
				else {
					evaluation.setAllFalse();
				}
				// prepare instance
				Instance instance = Instance(formula, evaluation, initialTemperatureFactor, minTemperature, cooldownFactor, equilibriumFactor, costFactor);
				// do the ANNEALING
				statesTested += instance.solve();
				solutionsCount += instance.getSolutionsCount();
				worstWeight += instance.getImprovementRatio();
				satisfiedTerms += static_cast<float>(instance.getSatisfiedTerms()) / formula.getTermsCount();
				if (!instance.foundBest()) {
					improvementTestBest += instance.getImprovementOfTestBest();
				}
				if (!instance.isCorrectSolution()) {
					++solutionNotFound;
				}
				// save the solution to a file - just once
				if (i == ITERATIONSCOUNT-1) {
					instance.printSolution(fout);
				}
				++iterations;
			}			
			fout.close();
		}

		assert(iterations > 0);

		// print out statistics
		cout << "Number of instances (each repeated " << ITERATIONSCOUNT << " times): " << inputFiles.size() << " --> " << iterations << " runs" << endl;
		cout << "--- average values ---" << endl;
		cout << "States tested: " << static_cast<float>(statesTested) / iterations << endl;
		cout << "Solutions found: " << static_cast<float>(solutionsCount) / iterations << endl;
		cout << "Satisfied terms ratio: " << satisfiedTerms * 100 / iterations << "%" << endl;
		cout << "(Best-Worst)/Worst weight: " << worstWeight * 100 / iterations << "%" << endl;
		cout << "Change of weight caused by TestBest: " << improvementTestBest / iterations << "%" << endl;
		cout << "Solution not found: " << static_cast<float>(solutionNotFound)*100 / iterations << "%" << endl;
	}

	return 0;
}